# Copyright (c) 2022, Miroslav Stoyanov & Weiwei Kong
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND
# IMPLIED. THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL
# ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE. THE USER ASSUMES
# RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING
# FROM OR ARISING OUT OF, IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.

from ctypes import c_char_p, c_int, c_double, c_void_p, POINTER, cdll, cast, CFUNCTYPE, Structure
from numpy.ctypeslib import as_ctypes
import numpy as np
import sys

from TasmanianConfig import __path_libdream__
from TasmanianConfig import TasmanianInputError as InputError

type_optim_obj_fn_single  = CFUNCTYPE(c_double, c_int, POINTER(c_double), POINTER(c_int))
type_optim_grad_fn_single = CFUNCTYPE(None, c_int, POINTER(c_double), POINTER(c_double), POINTER(c_int))
type_optim_proj_fn_single = CFUNCTYPE(None, c_int, POINTER(c_double), POINTER(c_double), POINTER(c_int))

# OptimizationStatus class (mirrors the C++ struct).
class optimization_status(Structure):
    pass
optimization_status._fields_ = [("performed_iterations", c_int), ("residual", c_double)]
def clean_status(input_status):
    return dict((field, getattr(input_status, field)) for field, _ in input_status._fields_)

pLibDTSG = cdll.LoadLibrary(__path_libdream__)

pLibDTSG.tsgGradientDescentState_Construct.restype = c_void_p
pLibDTSG.tsgGradientDescentState_Construct.argtypes = [c_int, POINTER(c_double), c_double]

pLibDTSG.tsgGradientDescentState_Destruct.argtypes = [c_void_p]

pLibDTSG.tsgGradientDescentState_GetNumDimensions.restype = c_int
pLibDTSG.tsgGradientDescentState_GetNumDimensions.argtypes = [c_void_p]
pLibDTSG.tsgGradientDescentState_GetAdaptiveStepsize.restype = c_double
pLibDTSG.tsgGradientDescentState_GetAdaptiveStepsize.argtypes = [c_void_p]
pLibDTSG.tsgGradientDescentState_GetX.argtypes = [c_void_p, POINTER(c_double)]

pLibDTSG.tsgGradientDescentState_SetAdaptiveStepsize.argtypes = [c_void_p, c_double]
pLibDTSG.tsgGradientDescentState_SetX.argtypes = [c_void_p, POINTER(c_double)]

pLibDTSG.tsgGradientDescent_AdaptProj.argtypes = [type_optim_obj_fn_single, type_optim_grad_fn_single, type_optim_proj_fn_single,
                                                  c_double, c_double, c_int, c_double, c_void_p, POINTER(c_int)]
pLibDTSG.tsgGradientDescent_AdaptProj.restype = optimization_status
pLibDTSG.tsgGradientDescent_Adapt.argtypes = [type_optim_obj_fn_single, type_optim_grad_fn_single, c_double, c_double, c_int,
                                              c_double, c_void_p, POINTER(c_int)]
pLibDTSG.tsgGradientDescent_Adapt.restype = optimization_status
pLibDTSG.tsgGradientDescent_Const.argtypes = [type_optim_grad_fn_single, c_double, c_int, c_double, c_void_p, POINTER(c_int)]
pLibDTSG.tsgGradientDescent_Const.restype = optimization_status

class GradientDescentState:
    '''
    Wrapper to class TasOptimization::GradientDescentState
    '''
    def __init__(self, aX0, fInitialStepsize):
        '''
        Constructs a new state with a given number of dimensions and particles.

        aX0              : a one-dimensional NumPy array representing the starting point.
        fInitialStepsize : initial adaptive stepsize.
        '''
        self.TasmanianGradientDescentState = True
        if (len(aX0.shape) != 1):
            raise InputError("aX0", "aX0 should be one-dimensional")
        self.pStatePntr = c_void_p(pLibDTSG.tsgGradientDescentState_Construct(aX0.shape[0], as_ctypes(aX0), fInitialStepsize))

    def __del__(self):
        '''
        Deletes an instance of the particle swarm state.
        '''
        pLibDTSG.tsgGradientDescentState_Destruct(self.pStatePntr)

    def getNumDimensions(self):
        '''
        Return the number of dimensions.
        '''
        return pLibDTSG.tsgGradientDescentState_GetNumDimensions(self.pStatePntr)

    def getAdaptiveStepsize(self):
        '''
        Return the adaptive stepsize.
        '''
        return pLibDTSG.tsgGradientDescentState_GetAdaptiveStepsize(self.pStatePntr)

    def getX(self):
        '''
        Return the current candidate point.
        '''
        iNumDims = self.getNumDimensions()
        aResult = np.zeros((iNumDims,), np.float64)
        pLibDTSG.tsgGradientDescentState_GetX(self.pStatePntr, as_ctypes(aResult))
        return aResult

    def setAdaptiveStepsize(self, fNewStepsize):
        '''
        Set new adaptive stepsize

        fNewStepsize : a double representing the new adaptive stepsize.
        '''
        pLibDTSG.tsgGradientDescentState_SetAdaptiveStepsize(self.pStatePntr, fNewStepsize)

    def setX(self, aXNew):
        '''
        Set new particle positions from a NumPy array.

        aXNew : a one-dimensional NumPy array representing the new candidate point.
        '''
        iNumDims = self.getNumDimensions()
        if (len(aXNew.shape) != 1):
            raise InputError("aXNew", "aXNew should be one-dimensional")
        if (aXNew.shape[0] != iNumDims):
            raise InputError("aXNew", "aXNew.shape[0] should match the number of dimensions")
        pLibDTSG.tsgGradientDescentState_SetX(self.pStatePntr, as_ctypes(aXNew))

# Helper methods for converting Python lambdas to C lambdas.
def convert_py_obj_fn_single(pObjectiveFunction, sCallerName):
    def cpp_obj_fn(num_dim, x_single_ptr, err_arr):
        err_arr[0] = 1
        aX = np.ctypeslib.as_array(x_single_ptr, (num_dim, ))
        aResult = pObjectiveFunction(aX)
        if not isinstance(aResult, float):
            print("ERROR: incorrect output from the objective function given to " + sCallerName +
                  "(), should be a 'float' but received '" + type(aResult).__name__ + "'")
            return 0
        err_arr[0] = 0
        return aResult
    return cpp_obj_fn

def convert_py_grad_fn_single(pGradientFunction, sCallerName):
    def cpp_grad_fn(num_dim, x_single_ptr, grad_ptr, err_arr):
        err_arr[0] = 1
        aX = np.ctypeslib.as_array(x_single_ptr, (num_dim,))
        aResult = pGradientFunction(aX)
        if aResult.shape != (num_dim, ):
            print("ERROR: incorrect output from the gradient function given to " + sCallerName +
                  "(), should be a NumPy array with shape (iNumDims,)")
            return
        aGrad = np.ctypeslib.as_array(grad_ptr, (num_dim,))
        aGrad[0:num_dim] = aResult[0:num_dim]
        err_arr[0] = 0
    return cpp_grad_fn

def convert_py_proj_fn_single(pProjectionFunction, sCallerName):
    def cpp_proj_fn(num_dim, x_single_ptr, proj_ptr, err_arr):
        err_arr[0] = 1
        aX = np.ctypeslib.as_array(x_single_ptr, (num_dim, ))
        aResult = pProjectionFunction(aX)
        if aResult.shape != (num_dim, ):
            print("ERROR: incorrect output from the projection function given to " + sCallerName +
                  "(), should be a NumPy array with shape (iNumDims,)")
            return
        aProj =  np.ctypeslib.as_array(proj_ptr, (num_dim,))
        aProj[0:num_dim] = aResult[0:num_dim]
        err_arr[0] = 0
    return cpp_proj_fn

# Main algorithms.
def AdaptiveProjectedGradientDescent(pObjectiveFunction, pGradientFunction, pProjectionFunction, fIncreaseCoeff, fDecreaseCoeff,
                                     iMaxIterations, fTolerance, oGradientDescentState):
    '''
    Wrapper around TasOptimization::GradientDescent().

    Runs at most iMaxIterations of the adaptive stepsize projected gradient descent algorithm on an input state oGradientDescentState
    to minimize the function pObjectiveFunction with gradient pGradientFunction and projection function pProjectionFunction. The
    parameters fIncreaseCoeff and fDecreaseCoeff control the evolution of the algorithm. The parameter fTolerance is a first-order
    stationarity tolerance used for early termination (see the C++ documentation for more details).

    Returns a dictionary containing the results of the run.

    pObjectiveFunction    : a Python lambda representing the objective function; it should take in a one-dimensional NumPy array
                            (x_single) and produce a double representing be the result of evaluating the objective function on
                            x_single; it is expected that x_single.shape[0] = .getNumDimensions().
    pGradientFunction     : a Python lambda representing the gradient of the objective function; it should take in a
                            one-dimensional NumPy array (x_single) and produce a one-dimensional NumPy representing the gradient
                            (grad) at x_single; it is expected that x_single.shape[0] = grad.shape[0] = .getNumDimensions().
    pProjectionFunction   : a Python lambda representing the domain projection function; it should take in a one-dimensional NumPy
                            array (x_single) and produce a one-dimensional NumPy representing the projection (proj) of x_single
                            onto the domain; it is expected that x_single.shape[0] = proj.shape[0] = .getNumDimensions().
    fIncreaseCoeff        : a positive double controlling how fast the stepsize is increased; it is expected to be greater than 1.
    fDecreaseCoeff        : a positive double controlling how fast the stepsize is decreased; it is expected to be greater than 1.
    iMaxIterations        : a positive integer representing the maximum number iterations the algorithm is run.
    fTolerance            : a positive double representing the stationarity tolerance for early tolerance.
    oGradientDescentState : an instance of the Python GradientDescentState class; it will contain the results of applying the algorithm.
    '''
    cpp_obj_fn = convert_py_obj_fn_single(pObjectiveFunction, "GradientDescent")
    cpp_grad_fn = convert_py_grad_fn_single(pGradientFunction, "GradientDescent")
    cpp_proj_fn = convert_py_proj_fn_single(pProjectionFunction, "GradientDescent")
    pErrorCode = (c_int * 1)()
    oStatus = pLibDTSG.tsgGradientDescent_AdaptProj(type_optim_obj_fn_single(cpp_obj_fn), type_optim_grad_fn_single(cpp_grad_fn),
                                                    type_optim_proj_fn_single(cpp_proj_fn), c_double(fIncreaseCoeff), c_double(fDecreaseCoeff),
                                                    c_int(iMaxIterations), c_double(fTolerance), oGradientDescentState.pStatePntr, pErrorCode)
    if pErrorCode[0] != 0:
        raise InputError("GradientDescent", "An error occurred during the call to Tasmanian.")
    return clean_status(oStatus)

def AdaptiveGradientDescent(pObjectiveFunction, pGradientFunction, fIncreaseCoeff, fDecreaseCoeff, iMaxIterations, fTolerance,
                            oGradientDescentState):
    '''
    Wrapper around TasOptimization::GradientDescent().

    Runs at most iMaxIterations of the adaptive stepsize (unconstrained) gradient descent algorithm on an input state oGradientDescentState
    to minimize the function pObjectiveFunction with gradient pGradientFunction and projection function pProjectionFunction. The
    parameters fIncreaseCoeff and fDecreaseCoeff control the evolution of the algorithm. The parameter fTolerance is a first-order
    stationarity tolerance used for early termination (see the C++ documentation for more details).

    Returns a dictionary containing the results of the run.

    pObjectiveFunction    : a Python lambda representing the objective function; it should take in a one-dimensional NumPy array
                            (x_single) and produce a double representing be the result of evaluating the objective function on
                            x_single; it is expected that x_single.shape[0] = .getNumDimensions().
    pGradientFunction     : a Python lambda representing the gradient of the objective function; it should take in a
                            one-dimensional NumPy array (x_single) and produce a one-dimensional NumPy representing the gradient
                            (grad) at x_single; it is expected that x_single.shape[0] = grad.shape[0] = .getNumDimensions().
    fIncreaseCoeff        : a positive double controlling how fast the stepsize is increased; it is expected to be greater than 1.
    fDecreaseCoeff        : a positive double controlling how fast the stepsize is decreased; it is expected to be greater than 1.
    iMaxIterations        : a positive integer representing the maximum number iterations the algorithm is run.
    fTolerance            : a positive double representing the stationarity tolerance for early tolerance.
    oGradientDescentState : an instance of the Python GradientDescentState class; it will contain the results of applying the algorithm.
    '''
    cpp_obj_fn = convert_py_obj_fn_single(pObjectiveFunction, "GradientDescent")
    cpp_grad_fn = convert_py_grad_fn_single(pGradientFunction, "GradientDescent")
    pErrorCode = (c_int * 1)()
    oStatus = pLibDTSG.tsgGradientDescent_Adapt(type_optim_obj_fn_single(cpp_obj_fn), type_optim_grad_fn_single(cpp_grad_fn),
                                                c_double(fIncreaseCoeff), c_double(fDecreaseCoeff), c_int(iMaxIterations),
                                                c_double(fTolerance), oGradientDescentState.pStatePntr, pErrorCode)
    if pErrorCode[0] != 0:
        raise InputError("GradientDescent", "An error occurred during the call to Tasmanian.")
    return clean_status(oStatus)

def GradientDescent(pGradientFunction, fStepsize, iMaxIterations, fTolerance, oGradientDescentState):
    '''
    Wrapper around TasOptimization::GradientDescent().

    Runs at most iMaxIterations of the constant stepsize (unconstrained) gradient descent algorithm on an input state oGradientDescentState
    to minimize the function pObjectiveFunction with gradient pGradientFunction and projection function pProjectionFunction.
    The parameter fTolerance is a first-order stationarity tolerance used for early termination (see the C++ documentation for
    more details).

    Returns a dictionary containing the results of the run.

    pGradientFunction     : a Python lambda representing the gradient of the objective function; it should take in a
                            one-dimensional NumPy array (x_single) and produce a one-dimensional NumPy representing the gradient
                            (grad) at x_single; it is expected that x_single.shape[0] = grad.shape[0] = .getNumDimensions().
    fStepsize             : a positive double representing the stepsize.
    iMaxIterations        : a positive integer representing the maximum number iterations the algorithm is run.
    fTolerance            : a positive double representing the stationarity tolerance for early tolerance.
    oGradientDescentState : an instance of the Python GradientDescentState class; it will contain the results of applying the algorithm.
    '''
    cpp_grad_fn = convert_py_grad_fn_single(pGradientFunction, "GradientDescent")
    pErrorCode = (c_int * 1)()
    oStatus = pLibDTSG.tsgGradientDescent_Const(type_optim_grad_fn_single(cpp_grad_fn), c_double(fStepsize), c_int(iMaxIterations),
                                                c_double(fTolerance), oGradientDescentState.pStatePntr, pErrorCode)
    if pErrorCode[0] != 0:
        raise InputError("GradientDescent", "An error occurred during the call to Tasmanian.")
    return clean_status(oStatus)




# Copyright (c) 2022, Miroslav Stoyanov & Weiwei Kong
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND
# IMPLIED. THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL
# ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE. THE USER ASSUMES
# RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING
# FROM OR ARISING OUT OF, IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.

import unittest
import Tasmanian
DREAM = Tasmanian.DREAM
Opt = Tasmanian.Optimization
import os
import numpy as np

import testCommon

ttc = testCommon.TestTasCommon()

class TestTasClass(unittest.TestCase):
    '''
    Tests for the Tasmanian Optimization python bindings module
    '''
    def __init__(self):
        unittest.TestCase.__init__(self, "testNothing")

    def testNothing(self):
        pass

    def checkParticleSwarmState(self):
        iNumDimensions = 2
        iNumParticles = 5
        # Initialization tests.
        state = Opt.ParticleSwarmState(iNumDimensions, iNumParticles)
        self.assertEqual(iNumDimensions, state.getNumDimensions())
        self.assertEqual(iNumParticles, state.getNumParticles())
        self.assertFalse(state.isPositionInitialized())
        self.assertFalse(state.isVelocityInitialized())
        self.assertFalse(state.isBestPositionInitialized())
        self.assertFalse(state.isCacheInitialized())
        # Loading/Unloading tests.
        pp = np.arange(iNumParticles * iNumDimensions, dtype=np.float64).reshape(iNumParticles, iNumDimensions)
        state.setParticlePositions(pp)
        self.assertTrue(state.isPositionInitialized())
        np.testing.assert_array_equal(pp, state.getParticlePositions())
        pv =  np.arange(iNumParticles * iNumDimensions, dtype=np.float64).reshape(iNumParticles, iNumDimensions)
        state.setParticleVelocities(pv)
        np.testing.assert_array_equal(pv, state.getParticleVelocities())
        bpp =  np.arange((iNumParticles+1) * iNumDimensions, dtype=np.float64).reshape(iNumParticles+1, iNumDimensions)
        state.setBestParticlePositions(bpp)
        self.assertTrue(state.isBestPositionInitialized())
        np.testing.assert_array_equal(bpp, state.getBestParticlePositions())
        state.clearBestParticles()
        self.assertFalse(state.isBestPositionInitialized())
        np.testing.assert_array_equal(np.zeros([iNumParticles + 1, iNumDimensions]), state.getBestParticlePositions())
        # Generation tests.
        state = Opt.ParticleSwarmState(iNumDimensions, iNumParticles)
        state.initializeParticlesInsideBox(np.array([-1.0, 1.0]), np.array([2.0, 3.0]),
                                           random01=DREAM.RandomGenerator(sType="default", iSeed=777))
        self.assertTrue(state.isPositionInitialized())
        self.assertTrue(state.isVelocityInitialized())
        pp = state.getParticlePositions()
        self.assertTrue(np.all((-1.0 <= pp[:,0]) & (pp[:,0] <= 2.0)))
        self.assertTrue(np.all((1.0 <= pp[:,1]) & (pp[:,1] <= 3.0)))

    def checkParticleSwarm(self):
        iNumDimensions = 2
        iNumParticles = 50
        state = Opt.ParticleSwarmState(iNumDimensions, iNumParticles)
        state.initializeParticlesInsideBox(np.array([-3.0, -2.0]), np.array([3.0, 2.0]),
                                           random01=DREAM.RandomGenerator(sType="default", iSeed=777))
        # Six-hump-camel function.
        shc = lambda x : (4 - 2.1*x[0]*x[0] + x[0]*x[0]*x[0]*x[0]/3)*x[0]*x[0] + x[0]*x[1] + (-4.0 + 4.0*x[1]*x[1])*x[1]*x[1]
        f = lambda x_batch : np.apply_along_axis(shc, 1, x_batch)
        inside = lambda x : bool((-3 <= x[0]) and (x[0] <= 3) and (-2 <= x[1]) and (x[1] <= 2))
        # Main call + tests.
        Opt.ParticleSwarm(f, inside, 0.5, 2, 2, 1, state, random01=DREAM.RandomGenerator(sType="default", iSeed=777))
        self.assertTrue(state.isCacheInitialized())
        state.clearCache()
        self.assertFalse(state.isCacheInitialized())
        iNumIterations = 200
        Opt.ParticleSwarm(f, inside, 0.5, 2, 2, iNumIterations, state)
        self.assertTrue(np.allclose(np.array([-0.08984201368301331, +0.7126564032704135]), state.getBestPosition()) or
                        np.allclose(np.array([+0.08984201368301331, -0.7126564032704135]), state.getBestPosition()) )

    def checkGradientDescentState(self):
        x0 = np.array([0.0, 1.0])
        stepsize = 2.0;
        # Initialization tests.
        state = Opt.GradientDescentState(x0, stepsize)
        self.assertEqual(len(x0), state.getNumDimensions())
        np.testing.assert_array_equal(x0, state.getX())
        self.assertEqual(stepsize, state.getAdaptiveStepsize())
        # Loading/Unloading tests.
        x1 = np.array([3.0, 4.0])
        state.setX(x1)
        np.testing.assert_array_equal(x1, state.getX())
        state.setAdaptiveStepsize(5.0)
        self.assertEqual(5.0, state.getAdaptiveStepsize())

    def checkGradientDescent(self):
        func = lambda x : 2.0 * x[0] * x[0] + x[1] * x[1] / 2.0
        grad = lambda x : np.array([4.0 * x[0], x[1]])
        lambda0 = 3.0
        xBar = np.array([0.0, 0.0])
        xBarConstr = np.array([-0.5, 0.5])
        tol = 1E-6
        iter_limit = 100
        # Main call + tests.
        state = Opt.GradientDescentState(np.array([-1.0, 1.0]), lambda0)
        result = Opt.GradientDescent(grad, 1/4.0, iter_limit, tol, state)
        self.assertLessEqual(result['performed_iterations'], iter_limit)
        self.assertTrue(np.allclose(xBar, state.getX(), atol=tol * 10))
        self.assertEqual(lambda0, state.getAdaptiveStepsize())
        state = Opt.GradientDescentState(np.array([-1.0, 1.0]), lambda0)
        result = Opt.AdaptiveGradientDescent(func, grad, 1.25, 1.25, iter_limit, tol, state)
        self.assertLessEqual(result['performed_iterations'], iter_limit)
        self.assertTrue(np.allclose(xBar, state.getX(), atol=tol * 10))
        proj = lambda x : np.array([min(max(-3.0, x[0]), -0.5), min(max(0.5, x[1]), 3.0)])
        state = Opt.GradientDescentState(np.array([-2.0, 2.0]), lambda0)
        result = Opt.AdaptiveProjectedGradientDescent(func, grad, proj, 1.25, 1.25, iter_limit, tol, state)
        self.assertLessEqual(result['performed_iterations'], iter_limit)
        self.assertTrue(np.allclose(xBarConstr, state.getX(), atol=tol))

    def performOptTests(self):
        self.checkParticleSwarmState()
        self.checkParticleSwarm()
        self.checkGradientDescentState()
        self.checkGradientDescent()

if __name__ == "__main__":
    tester = TestTasClass()
    tester.performOptTests()



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

import numpy as np
from random import uniform
import Tasmanian

def example_11():

    print("\n---------------------------------------------------------------------------------------------------\n")
    print("Example 11: Using unstructured data\n")

    iNumInputs = 2 # using two inputs for testing

    # test the error on a uniform dense grid with 10K points
    iTestGridSize = 100
    dx = np.linspace(-1.0, 1.0, iTestGridSize) # sample on a uniform grid
    aMeshX, aMeshY = np.meshgrid(dx, dx)
    aTestPoints = np.column_stack([aMeshX.reshape((iTestGridSize**2, 1)),
                                   aMeshY.reshape((iTestGridSize**2, 1))])

    def get_error(grid, model, aTestPoints):
        aGridResult = grid.evaluateBatch(aTestPoints)
        aModelResult = np.empty((aTestPoints.shape[0], 1), np.float64)
        for i in range(aTestPoints.shape[0]):
            aModelResult[i,:] = model(aTestPoints[i,:])
        return np.max(np.abs(aModelResult[:,0] - aGridResult[:,0]))

    def model(aX):
        return np.ones((1,)) * np.exp( - aX[0]**2 - aX[1]**2 )

    grid = Tasmanian.makeGlobalGrid(iNumInputs, 1, 4, "level", "clenshaw-curtis")

    # generate random data
    iNumData = 2000
    inputs = np.zeros((iNumData, iNumInputs))
    outputs = np.zeros((iNumData, 1))

    for i in range(iNumData):
        inputs[i,0] = uniform(-1.0, 1.0)
        inputs[i,1] = uniform(-1.0, 1.0)
        outputs[i,:] = model(inputs[i,:])

    if (not grid.isAccelerationAvailable("cpu-blas") and
        not grid.isAccelerationAvailable("gpu-cuda") and
        not grid.isAccelerationAvailable("gpu-magma")):
        print("Skipping example 11, BLAS, CUDA, or MAGMA acceleration required.")
        return

    Tasmanian.loadUnstructuredDataL2(inputs, outputs, 1.E-4, grid)

    print("Using construction from unstructured (random) data")
    print("    approximatino error = {0:1.6E}".format(get_error(grid, model, aTestPoints)))
    print("  note: compared to the C++ code, this example does not fix the random seed")
    print("        the error will be slightly different every time you run the example")
    print("")


if (__name__ == "__main__"):
    example_11()



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

from Tasmanian import Optimization as Opt, makeGlobalGrid
import numpy as np

def example_02():
    print("\n---------------------------------------------------------------------------------------------------\n")
    print("EXAMPLE 2: use various Gradient Descent algorithms to minimize a convex quadratic")

    # Define the objective function `func` and and its gradient `grad`.
    func = lambda x : 2.0 * (x[0] - 1.0) * (x[0] - 1.0) + (x[1] - 2.0) * (x[1] - 2.0) / 2.0
    grad = lambda x : np.array([4.0 * (x[0] - 1.0), x[1] - 2.0])

    # Create an empty Gradient Descent State.
    x0 = np.array([0.0, 0.0])
    state = Opt.GradientDescentState(x0, 0.0)

    # Run the Gradient Descent (GD) algorithm and check the output. Note that the adaptive stepsize in the state is NOT used.
    iNumIterations = 200;
    fTolerance = 1E-3
    dInfo = Opt.GradientDescent(grad, 1/8.0, iNumIterations, fTolerance, state)

    aGlobal = np.array([1.0, 2.0])
    aSolution = state.getX()

    sResult = ""
    fL2Error = 0.0
    for i in range(2):
        sResult = "{0:1s}{1:13.5f}".format(sResult, aSolution[i])
        fL2Error += (aSolution[i] - aGlobal[i]) ** 2
    fL2Error = np.sqrt(fL2Error)
    print("\nUsing the Gradient Descent algorithm on the EXACT objective function, the computed solution is:")
    print(" iterations: {0:14}".format(dInfo['performed_iterations']))
    print("   computed: {0:1s}".format(sResult))
    print("   L2 error: {0:14e}".format(fL2Error))

    # Create a surrogate model for the quadratic.
    # NOTE: local polynomial grids are not well suited for first-order methods like gradient descent.
    grid = makeGlobalGrid(2, 1, 2, "iptotal", "leja")
    needed_points = grid.getNeededPoints()
    needed_values = np.resize(np.apply_along_axis(func, 1, needed_points), [grid.getNumNeeded(), 1])
    grid.loadNeededValues(needed_values)
    surrogate_func = lambda x : grid.evaluate(x)
    surrogate_grad = lambda x : grid.differentiate(x)

    # Run the GD algorithm on the surrogate and check the output.
    state = Opt.GradientDescentState(x0, 0.0)
    dInfo = Opt.GradientDescent(surrogate_grad, 1/8.0, iNumIterations, fTolerance, state)

    aSolution = state.getX()

    sResult = ""
    fL2Error = 0.0
    for i in range(2):
        sResult = "{0:1s}{1:13.5f}".format(sResult, aSolution[i])
        fL2Error += (aSolution[i] - aGlobal[i]) ** 2
    fL2Error = np.sqrt(fL2Error)
    print("\nUsing the Gradient Descent algorithm on the SURROGATE objective function, the computed solution is:")
    print(" iterations: {0:14}".format(dInfo['performed_iterations']))
    print("   computed: {0:1s}".format(sResult))
    print("   L2 error: {0:14e}".format(fL2Error))

    # Run the adaptive GD algorithm and check the output. Note that the adaptive stepsize in the state IS used here.
    state = Opt.GradientDescentState(x0, 1.0)
    dInfo = Opt.AdaptiveGradientDescent(func, grad, 1.25, 1.25, iNumIterations, fTolerance, state)

    aSolution = state.getX()

    sResult = ""
    fL2Error = 0.0
    for i in range(2):
        sResult = "{0:1s}{1:13.5f}".format(sResult, aSolution[i])
        fL2Error += (aSolution[i] - aGlobal[i]) ** 2
    fL2Error = np.sqrt(fL2Error)
    print("\nUsing the ADAPTIVE Gradient Descent algorithm on the EXACT objective function, the computed solution is:")
    print(" iterations: {0:14}".format(dInfo['performed_iterations']))
    print("   computed: {0:1s}".format(sResult))
    print("   L2 error: {0:14e}".format(fL2Error))

    # Run the projected adaptive GD algorithm and check the output. The domain constraint is implicitly enforced by the
    # projection function `proj`.
    state = Opt.GradientDescentState(x0, 1.0)
    proj = lambda x : np.array([min(0.5, max(0.0, x[0])),
                                min(1.5, max(0.0, x[1]))]) # projection onto the box [0.0, 0.5] x [0.0, 1.5]
    dInfo = Opt.AdaptiveProjectedGradientDescent(func, grad, proj, 1.25, 1.25, iNumIterations, fTolerance, state)

    aGlobal = np.array([0.5, 1.5])
    aSolution = state.getX()

    sResult = ""
    fL2Error = 0.0
    for i in range(2):
        sResult = "{0:1s}{1:13.5f}".format(sResult, aSolution[i])
        fL2Error += (aSolution[i] - aGlobal[i]) ** 2
    fL2Error = np.sqrt(fL2Error)
    print("\nUsing the ADAPTIVE PROJECTED Gradient Descent algorithm on the EXACT objective function, the computed solution is:")
    print(" iterations: {0:14}".format(dInfo['performed_iterations']))
    print("   computed: {0:1s}".format(sResult))
    print("   L2 error: {0:14e}".format(fL2Error))


if __name__ == "__main__":
    example_02()



# Copyright (c) 2022, Miroslav Stoyanov & Weiwei Kong
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND
# IMPLIED. THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF
# THE SOFTWARE WILL NOT INFRINGE ANY PATENT, COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL
# ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE. THE USER ASSUMES
# RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING
# FROM OR ARISING OUT OF, IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.

from ctypes import c_char_p, c_int, c_double, c_void_p, POINTER, cdll, cast, CFUNCTYPE
from numpy.ctypeslib import as_ctypes
import numpy as np
import sys

from TasmanianConfig import __path_libdream__
from TasmanianConfig import TasmanianInputError as InputError

import TasmanianDREAM as DREAM

type_optim_obj_fn = CFUNCTYPE(None, c_int, c_int, POINTER(c_double), POINTER(c_double), POINTER(c_int))
type_optim_dom_fn = CFUNCTYPE(c_int, c_int, POINTER(c_double), POINTER(c_int))
type_dream_random = CFUNCTYPE(c_double)

pLibDTSG = cdll.LoadLibrary(__path_libdream__)

pLibDTSG.tsgParticleSwarmState_Construct.restype = c_void_p
pLibDTSG.tsgParticleSwarmState_Construct.argtypes = [c_int, c_int]

pLibDTSG.tsgParticleSwarmState_Destruct.argtypes = [c_void_p]

pLibDTSG.tsgParticleSwarmState_GetNumDimensions.restype = c_int
pLibDTSG.tsgParticleSwarmState_GetNumDimensions.argtype = [c_void_p]
pLibDTSG.tsgParticleSwarmState_GetNumParticles.restype = c_int
pLibDTSG.tsgParticleSwarmState_GetNumParticles.argtype = [c_void_p]
pLibDTSG.tsgParticleSwarmState_GetParticlePositions.argtypes = [c_void_p, POINTER(c_double)]
pLibDTSG.tsgParticleSwarmState_GetParticleVelocities.argtypes = [c_void_p, POINTER(c_double)]
pLibDTSG.tsgParticleSwarmState_GetBestParticlePositions.argtypes = [c_void_p, POINTER(c_double)]
pLibDTSG.tsgParticleSwarmState_GetBestPosition.argtypes = [c_void_p, POINTER(c_double)]

pLibDTSG.tsgParticleSwarmState_IsPositionInitialized.restype = c_int
pLibDTSG.tsgParticleSwarmState_IsPositionInitialized.argtype = [c_void_p]
pLibDTSG.tsgParticleSwarmState_IsVelocityInitialized.restype = c_int
pLibDTSG.tsgParticleSwarmState_IsVelocityInitialized.argtype = [c_void_p]
pLibDTSG.tsgParticleSwarmState_IsBestPositionInitialized.restype = c_int
pLibDTSG.tsgParticleSwarmState_IsBestPositionInitialized.argtype = [c_void_p]
pLibDTSG.tsgParticleSwarmState_IsCacheInitialized.restype = c_int
pLibDTSG.tsgParticleSwarmState_IsCacheInitialized.argtype = [c_void_p]

pLibDTSG.tsgParticleSwarmState_SetParticlePositions.argtypes = [c_void_p, POINTER(c_double)]
pLibDTSG.tsgParticleSwarmState_SetParticleVelocities.argtypes = [c_void_p, POINTER(c_double)]
pLibDTSG.tsgParticleSwarmState_SetBestParticlePositions.argtypes = [c_void_p, POINTER(c_double)]

pLibDTSG.tsgParticleSwarmState_ClearBestParticles.argtypes = [c_void_p]
pLibDTSG.tsgParticleSwarmState_ClearCache.argtypes = [c_void_p]
pLibDTSG.tsgParticleSwarmState_InitializeParticlesInsideBox.argtypes = [c_void_p, POINTER(c_double), POINTER(c_double), c_char_p, c_int, type_dream_random]
pLibDTSG.tsgParticleSwarm.argtypes = [type_optim_obj_fn, type_optim_dom_fn, c_double, c_double, c_double, c_int, c_void_p,
                                      c_char_p, c_int, type_dream_random, POINTER(c_int)]

class ParticleSwarmState:
    '''
    Wrapper to class TasOptimization::ParticleSwarmState
    '''
    def __init__(self, iNumDimensions, iNumParticles):
        '''
        Constructs a new state with a given number of dimensions and particles.

        iNumDimensions : positive integer indicating the number of dimensions.
        iNumParticles  : positive integer indicating the number of particles.
        '''
        self.TasmanianParticleSwarmState = True
        self.pStatePntr = c_void_p(pLibDTSG.tsgParticleSwarmState_Construct(iNumDimensions, iNumParticles))

    def __del__(self):
        '''
        Deletes an instance of the particle swarm state.
        '''
        pLibDTSG.tsgParticleSwarmState_Destruct(self.pStatePntr)

    def getNumDimensions(self):
        '''
        Return the number of dimensions.
        '''
        return pLibDTSG.tsgParticleSwarmState_GetNumDimensions(self.pStatePntr)

    def getNumParticles(self):
        '''
        Return the number of particles.
        '''
        return pLibDTSG.tsgParticleSwarmState_GetNumParticles(self.pStatePntr)

    def getParticlePositions(self):
        '''
        Return the particle positions as a 2D NumPy array. The shape of this array is (self.getNumParticles(), self.getNumDimensions())
        and the i-th row if this array corresponds to the position of the i-th particle.
        '''
        iNumDims = self.getNumDimensions()
        iNumPart = self.getNumParticles()
        aResult = np.zeros((iNumDims * iNumPart,), np.float64)
        pLibDTSG.tsgParticleSwarmState_GetParticlePositions(self.pStatePntr, as_ctypes(aResult))
        return aResult.reshape((iNumPart, iNumDims))

    def getParticleVelocities(self):
        '''
        Return the particle velocities as a 2D NumPy array. The shape of this array is (self.getNumParticles(), self.getNumDimensions())
        and the i-th row if this array corresponds to the velocity of the i-th particle.
        '''
        iNumDims = self.getNumDimensions()
        iNumPart = self.getNumParticles()
        aResult = np.zeros((iNumDims * iNumPart,), np.float64)
        pLibDTSG.tsgParticleSwarmState_GetParticleVelocities(self.pStatePntr, as_ctypes(aResult))
        return aResult.reshape((iNumPart, iNumDims))

    def getBestParticlePositions(self):
        '''
        Return the best particle positions as a 2D NumPy array.  The shape of this array is (self.getNumParticles()+1 , self.getNumDimensions())
        and the i-th row if this array corresponds to the best position of the i-th particle for the first .getNumParticles()
        particles. The last row corresponds to the best particle position of the swarm.
        '''
        iNumDims = self.getNumDimensions()
        iNumPart = self.getNumParticles()
        aResult = np.zeros((iNumDims * (iNumPart + 1),), np.float64)
        pLibDTSG.tsgParticleSwarmState_GetBestParticlePositions(self.pStatePntr, as_ctypes(aResult))
        return aResult.reshape((iNumPart + 1, iNumDims))

    def getBestPosition(self):
        '''
        Return the best particle position of the swarm as a 1D NumPy array. The size of the array is self.getNumDimensions().
        '''
        iNumDims = self.getNumDimensions()
        aResult = np.zeros((iNumDims,), np.float64)
        pLibDTSG.tsgParticleSwarmState_GetBestPosition(self.pStatePntr, as_ctypes(aResult))
        return aResult

    def isPositionInitialized(self):
        '''
        Returns True if the particle positions have been initialized and False otherwise.
        '''
        return bool(pLibDTSG.tsgParticleSwarmState_IsPositionInitialized(self.pStatePntr))

    def isVelocityInitialized(self):
        '''
        Returns True if the particle velocities have been initialized and False otherwise.
        '''
        return bool(pLibDTSG.tsgParticleSwarmState_IsVelocityInitialized(self.pStatePntr))

    def isBestPositionInitialized(self):
        '''
        Returns True if the best particle positions have been initialized and False otherwise.
        '''
        return bool(pLibDTSG.tsgParticleSwarmState_IsBestPositionInitialized(self.pStatePntr))

    def isCacheInitialized(self):
        '''
        Returns True if the cache has been initialized and False otherwise.
        '''
        return bool(pLibDTSG.tsgParticleSwarmState_IsCacheInitialized(self.pStatePntr))

    def setParticlePositions(self, llfNewPPosns):
        '''
        Set new particle positions from a NumPy array.

        llfNewPPosns : a two-dimensional numpy.ndarray with
            llfNewPPosns.shape[0] = self.getNumParticles()
            llfNewPPosns.shape[1] = self.getNumDimensions()
        '''
        iNumPart = self.getNumParticles()
        iNumDims = self.getNumDimensions()
        if (llfNewPPosns.shape[0] != iNumPart):
            raise InputError("llfNewPPosns", "llfNewPPosns.shape[0] should match the number of particles")
        if (llfNewPPosns.shape[1] != iNumDims):
            raise InputError("llfNewPPosns", "llfNewPPosns.shape[1] should match the number of dimensions")
        llfNewPPosns.resize((iNumDims * iNumPart,))
        pLibDTSG.tsgParticleSwarmState_SetParticlePositions(self.pStatePntr, as_ctypes(llfNewPPosns))
        llfNewPPosns.resize((iNumPart, iNumDims))

    def setParticleVelocities(self, llfNewPVelcs):
        '''
        Set new particle velocities from a NumPy array.

        llfNewPVelcs : a two-dimensional numpy.ndarray with
            llfNewPVelcs.shape[0] = self.getNumParticles()
            llfNewPVelcs.shape[1] = self.getNumDimensions()
        '''
        iNumPart = self.getNumParticles()
        iNumDims = self.getNumDimensions()
        if (llfNewPVelcs.shape[0] != iNumPart):
            raise InputError("llfNewPVelcs", "llfNewPVelcs.shape[0] should match the number of particles")
        if (llfNewPVelcs.shape[1] != iNumDims):
            raise InputError("llfNewPVelcs", "llfNewPVelcs.shape[1] should match the number of dimensions")
        llfNewPVelcs.resize((iNumDims * iNumPart,))
        pLibDTSG.tsgParticleSwarmState_SetParticleVelocities(self.pStatePntr, as_ctypes(llfNewPVelcs))
        llfNewPVelcs.resize((iNumPart, iNumDims))

    def setBestParticlePositions(self, llfNewBPPosns):
        '''
        Set new best particle positions from a NumPy array.

        llfNewBPPosns : a two-dimensional numpy.ndarray with
            llfNewPVelcs.shape[0] = self.getNumParticles() + 1
            llfNewPVelcs.shape[1] = self.getNumDimensions()
        '''
        iNumPart = self.getNumParticles()
        iNumDims = self.getNumDimensions()
        if (llfNewBPPosns.shape[0] != iNumPart + 1):
            raise InputError("llfNewBPPosns", "llfNewBPPosns.shape[0] should match the number of particles + 1")
        if (llfNewBPPosns.shape[1] != iNumDims):
            raise InputError("llfNewBPPosns", "llfNewBPPosns.shape[1] should match the number of dimensions")
        llfNewBPPosns.resize(((iNumPart + 1) * iNumDims,))
        pLibDTSG.tsgParticleSwarmState_SetBestParticlePositions(self.pStatePntr, as_ctypes(llfNewBPPosns))
        llfNewBPPosns.resize((iNumPart + 1, iNumDims))

    def clearBestParticles(self):
        '''
        Clears the vector of best particles.
        '''
        pLibDTSG.tsgParticleSwarmState_ClearBestParticles(self.pStatePntr)

    def clearCache(self):
        '''
        Clears the internal cache of the managed ParticleSwarm C++ object.
        '''
        pLibDTSG.tsgParticleSwarmState_ClearCache(self.pStatePntr)

    def initializeParticlesInsideBox(self, lfBoxLower, lfBoxUpper, random01 = DREAM.RandomGenerator(sType = "default")):
        '''
        Initialize the particle swarm state with randomized particles (determined by gen_random01) inside a box bounded by the
        parameters box_lower and box_upper. The i-th entry in these parameters respectively represent the lower and upper
        bounds on the particle position values for the i-th dimension. The parameter random01 controls the distribution of
        the particles.

        lfBoxLower : a one-dimensional NumPy array with lfBoxLower.shape[0] = self.getNumDimensions()
        lfBoxUpper : a one-dimensional NumPy array with lfBoxLower.shape[0] = self.getNumDimensions()
        random01   : a DREAM.RandomGenerator instance that produces floats in [0,1]
        '''
        iNumDims = self.getNumDimensions()
        if (lfBoxLower.shape != (iNumDims,)):
            raise InputError("lfBoxLower", "lfBoxLower.shape should be (self.getNumDimensions(),)")
        if (lfBoxUpper.shape != (iNumDims,)):
            raise InputError("lfBoxUpper", "lfBoxUpper.shape should be (self.getNumDimensions(),)")
        pLibDTSG.tsgParticleSwarmState_InitializeParticlesInsideBox(self.pStatePntr, as_ctypes(lfBoxLower), as_ctypes(lfBoxUpper),
                                                                    c_char_p(random01.sType), c_int(random01.iSeed),
                                                                    type_dream_random(random01.pCallable))

def ParticleSwarm(pObjectiveFunction, pInside, fInertiaWeight, fCognitiveCoeff, fSocialCoeff, iNumIterations, oParticleSwarmState,
                  random01 = DREAM.RandomGenerator(sType = "default")):
    '''
    Wrapper around TasOptimization::ParticleSwarm().

    Runs iNumIterations of the particle swarm algorithm on an input state cParticleSwarmState to minimize the function
    pObjectiveFunction over a domain specified by pInside. The parameters fInertiaWeight, fCognitiveCoeff, and fSocialCoeff
    control the evolution of the algorithm. The parameter random01 controls the distribution of the particles.

    pObjectiveFunction  : a Python lambda representing the objective function; it should take in one 2D NumPy array (x_batch)
                          and produce a 1D NumPy array, sized (iNumBatch,), whose i-th entry should be the result of evaluating
                          the objective function to x_batch[i,:]; it is expected that x_batch.shape[1] = .getNumDimensions().
    pInside             : a Python lambda representing the function domain; it should take in a 1D NumPy array (x) and produce a
                          Boolean (isInside); when called, it should return True if x is in the domain and False otherwise;
                          it is expected that x.shape[0] = .getNumDimensions().
    fInertiaWeight      : a double that controls the speed of the particles; should be in (0,1).
    fCognitiveCoeff     : a double that controls how much a particle favors its own trajectory; usually in [1,3].
    fSocialCoeff        : a double that controls how much a particle favors the swarm's trajectories; usually in [1,3].
    iNumIterations      : a positive integer representing the number iterations the algorithm is run.
    oParticleSwarmState : an instance of the Python ParticleSwarmState class; it will contain the results of applying the algorithm.
    random01            : (optional) a DREAM.RandomGenerator instance that produces floats in [0,1]
    '''
    def cpp_obj_fn(num_dim, num_batch, x_batch_ptr, fval_ptr, err_arr):
        err_arr[0] = 1
        aX = np.ctypeslib.as_array(x_batch_ptr, (num_batch, num_dim))
        aResult = pObjectiveFunction(aX)
        if aResult.shape != (num_batch, ):
            print("ERROR: incorrect output from the objective function given to ParticleSwarm(), should be a NumPy array with shape (iNumBatch,)")
            return
        aFVal = np.ctypeslib.as_array(fval_ptr, (num_batch,))
        aFVal[0:num_batch] = aResult[0:num_batch]
        err_arr[0] = 0

    def cpp_dom_fn(num_dim, x_ptr, err_arr):
        err_arr[0] = 1
        aX = np.ctypeslib.as_array(x_ptr, (num_dim,))
        iResult = pInside(aX)
        if not isinstance(iResult, bool):
            print("ERROR: incorrect output from the domain function given to ParticleSwarm(), should be a 'bool' but received '" +
                  type(iResult).__name__ + "'")
            return False
        err_arr[0] = 0
        return iResult

    pErrorCode = (c_int * 1)()
    pLibDTSG.tsgParticleSwarm(type_optim_obj_fn(cpp_obj_fn), type_optim_dom_fn(cpp_dom_fn), c_double(fInertiaWeight),
                              c_double(fCognitiveCoeff), c_double(fSocialCoeff), c_int(iNumIterations),
                              oParticleSwarmState.pStatePntr, c_char_p(random01.sType), c_int(random01.iSeed),
                              type_dream_random(random01.pCallable), pErrorCode)

    if pErrorCode[0] != 0:
        raise InputError("ParticleSwarm", "An error occurred during the call to Tasmanian.")




##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

from Tasmanian import Optimization as Opt, DREAM, makeLocalPolynomialGrid
import numpy as np

def example_01():
    print("\n---------------------------------------------------------------------------------------------------\n")
    print("EXAMPLE 1: use the Particle Swarm algorithm to minimize the six-hump camel function")

    # Create an empty Particle Swarm state.
    iNumDimensions = 2
    iNumParticles = 50
    randUnif = DREAM.RandomGenerator(sType="default", iSeed=777)
    state = Opt.ParticleSwarmState(iNumDimensions, iNumParticles)

    # Load the state with uniformly initialized particles in the domain [-3.0, 3.0] x [-2.0, 2.0].
    state.initializeParticlesInsideBox(np.array([-3.0, -2.0]), np.array([3.0, 2.0]), random01=randUnif)

    # Define the batched objective function `func` and a function `inside` that returns True if its input is inside the domain.
    # NOTE: `shc` is the unbatched six-hump camel function.
    shc = lambda x : (4 - 2.1*x[0]*x[0] + x[0]*x[0]*x[0]*x[0]/3)*x[0]*x[0] + x[0]*x[1] + (-4.0 + 4.0*x[1]*x[1])*x[1]*x[1]
    func = lambda x_batch : np.apply_along_axis(shc, 1, x_batch)
    inside = lambda x : bool((-3 <= x[0]) and (x[0] <= 3) and (-2 <= x[1]) and (x[1] <= 2))

    # Run the Particle Swarm (PS) algorithm and check the output.
    iNumIterations = 200;
    Opt.ParticleSwarm(func, inside, 0.5, 2, 2, iNumIterations, state, random01=randUnif)

    aLocal1 = np.array([-0.08984201368301331, +0.7126564032704135])
    aLocal2 = np.array([+0.08984201368301331, -0.7126564032704135])
    aSolution = state.getBestPosition()

    sResult = ""
    fL2Error = 0.0
    for i in range(iNumDimensions):
        sResult = "{0:1s}{1:13.5f}".format(sResult, aSolution[i])
        fL2Error += min((aSolution[i] - aLocal1[i]) ** 2,
                        (aSolution[i] - aLocal2[i]) ** 2)
    fL2Error = np.sqrt(fL2Error)
    print("\nUsing the Particle Swarm algorithm on the EXACT objective function, the computed solution is:")
    print(" computed: {0:1s}".format(sResult))
    print(" L2 error: {0:14e}".format(fL2Error))

    # Create a surrogate model for the six-hump camel function.
    grid = makeLocalPolynomialGrid(2, 1, 10, iOrder=1)
    grid.setDomainTransform(np.array([[-3.0, 3.0], [-2.0, 2.0]])) # set the non-canonical domain
    needed_points = grid.getNeededPoints()
    needed_values = np.resize(func(needed_points), [grid.getNumNeeded(), 1])
    grid.loadNeededValues(needed_values)
    surrogate_func = lambda x_batch : np.squeeze(grid.evaluateBatch(x_batch))

    # Run the PS algorithm on the surrogate and check the output.
    state = Opt.ParticleSwarmState(iNumDimensions, iNumParticles)
    state.initializeParticlesInsideBox(np.array([-3.0, -2.0]), np.array([3.0, 2.0]), random01=randUnif)
    Opt.ParticleSwarm(surrogate_func, inside, 0.5, 2, 2, iNumIterations, state, random01=randUnif)

    aSolution = state.getBestPosition()

    sResult = ""
    fL2Error = 0.0
    for i in range(iNumDimensions):
        sResult = "{0:1s}{1:13.5f}".format(sResult, aSolution[i])
        fL2Error += min((aSolution[i] - aLocal1[i]) ** 2,
                        (aSolution[i] - aLocal2[i]) ** 2)
    fL2Error = np.sqrt(fL2Error)
    print("\nUsing the Particle Swarm algorithm on the SURROGATE objective function, the computed solution is:")
    print(" computed: {0:1s}".format(sResult))
    print(" L2 error: {0:14e}".format(fL2Error))

if __name__ == "__main__":
    example_01()
